@@ -209,10 +209,9 @@ GList*     gtk_menu_get_for_attach_widget (GtkWidget           *widget);
 #define gtk_menu_insert(menu,child,pos)	gtk_menu_shell_insert ((GtkMenuShell *)(menu),(child),(pos))
 #endif /* GTK_DISABLE_DEPRECATED */
 
-/* private */
-
-void      _gtk_menu_set_reserve_toggle_size (GtkMenu  *menu,
-                                             gboolean  reserve);
+void     gtk_menu_set_reserve_toggle_size (GtkMenu  *menu,
+                                          gboolean   reserve_toggle_size);
+gboolean gtk_menu_get_reserve_toggle_size (GtkMenu  *menu);
 
 
 G_END_DECLS
